#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>

#include <asm/ioctls.h>
#include <sys/ioctl.h>

/* PSEUDO VBI headers */
#include "psvbi/decoder/psvbi.h"
#include "psvbi/decoder/psvbid-plugins.h"

/* VTR driver */
#include "vtr/vtr.h"
#include "vtr/vtr-ldisc.h"
#include "vtr/vtr-dev.h"
#include "vtr/vtr-cmd.h"

/* VTR workouts */
#define DEV_SERIAL "/dev/ttyS0"
#define DEV_VTR	   "/home/verem/vtr/vtr%d"
#define DEV_VTR_ALIAS 1
static int fd_serial = 0;
static int ldisc = N_VTR;
static int fd_vtr = 0, vtr_num;
static char vtr_device[128];


#ifdef USE_SYSLOG
#include <syslog.h>
#define LOG_EVENT syslog(LOG_NOTICE, 
#warning syslog used
#else
#define LOG_EVENT fprintf(stderr,
#warning stdout used
#endif /* USE_SYSLOG */


void __attribute__ ((constructor)) my_init(void)
{
    int err;
    
#ifdef USE_SYSLOG
    openlog("psvbid::test2",LOG_PID | LOG_CONS,LOG_DAEMON);
#endif

    // open device
    if((fd_serial = open(DEV_SERIAL, O_RDWR))<0)
    {
	LOG_EVENT "Unable to open device: \"%s\"\n", DEV_SERIAL);
	return;
    }
    else
    {
	LOG_EVENT "Device \"%s\" opened successfuly\n", DEV_SERIAL);
    };
				
    // change line disc to VTR style:
    if((err = ioctl(fd_serial, TIOCSETD, &ldisc)) != 0)
    {
        LOG_EVENT "Unable to changle line disc, [error_code: %d]\n ", errno);
        return ;
    };

    // request VTR device number			    
    if((err = ioctl(fd_serial, VTR_MAGIC, vtr_num = DEV_VTR_ALIAS)) != 0)
    {
        LOG_EVENT "Unable set vtr device alias number [error_code: %d]\n ", errno);
	return;
    }
    else
    {
	sprintf(vtr_device, DEV_VTR ,vtr_num);
	LOG_EVENT "Use '%s' for accesing to VTR\n",vtr_device);
    };

    // open VTR
    if((fd_vtr = open(vtr_device, O_RDWR))<0)
    {
        LOG_EVENT "Unable to open '%s' [error_code: %d]\n ", vtr_device, errno);
	return;
    };
										
    LOG_EVENT "VTR%d='%s' opened and ready\n ", vtr_num, vtr_device);
};

void __attribute__ ((destructor)) my_fini(void)
{
    LOG_EVENT "exiting....\n");
    
    // close devs
    if(fd_serial>0)
    {
	if(fd_vtr>0)
	    close(fd_vtr);
	    
	close(fd_serial);
    };

    LOG_EVENT "exit done!\n");

};


static int detect_ok(struct psvbi_block* block)
{
    char buf[1024];
    struct vtr_get_timed vtr_data;
    long current;
    
    
    /* request recent 'corrected LTC' */
    ioctl(fd_vtr, VTR_DEV_IOCTL_CMD(VTR_DEV_GET_TC, 0), &vtr_data);
    
    /* request kernel juffies */
    ioctl(fd_vtr, VTR_DEV_IOCTL_CMD(VTR_DEV_GET_JIFFIES, 0), &current);
    
    // create dump of data
    sprintf
    (
        buf,
/* text form */

	"FOUND BLOCK !\n"
	"\tVTR   -> LTC=%.8X, LTC_TIME=%ld, KRN_TIME=%ld, KRN_TIME-LTC_TIME=%ld\n"
        "\tPSVBI -> CORRECTIONS=%d,POS=[%d,%d], DATA=[%.2X %.2X %.2X %.2X %.2X %.2X %.2X %.2X %.2X %.2X %.2X %.2X %.2X %.2X]",

/* vtr */
	vtr_data.data.corrected_ltc,
	vtr_data.time,
	current,
	current - vtr_data.time,
	
/* psvbi */
        block->corrections,
        block->_h,block->_v,
        (unsigned int)block->decoded[0],(unsigned int)block->decoded[1],(unsigned int)block->decoded[2],
        (unsigned int)block->decoded[3],(unsigned int)block->decoded[4],(unsigned int)block->decoded[5],
        (unsigned int)block->decoded[6],(unsigned int)block->decoded[7],(unsigned int)block->decoded[8],
        (unsigned int)block->decoded[9],(unsigned int)block->decoded[10],(unsigned int)block->decoded[11],
        (unsigned int)block->decoded[12],(unsigned int)block->decoded[13]
    );
 
#ifdef USE_SYSLOG
    syslog(LOG_NOTICE, buf);
#else
    printf("test1::detect_ok %s\n",buf);
#endif	/* USE_SYSLOG */
    
    return 0;
};


static int detect_fail(struct psvbi_block* block)
{
    return 0;
};


/* ---------------------------------------------------------------------------

    Define plugin attributest
    
----------------------------------------------------------------------------*/

struct psvbid_plugin PLUGIN = 
{
    .name = "test1",
    .ok = detect_ok,
    .fail = NULL	// detect_fail
};

