#include <string.h>
#include <stdio.h>

#include "psvbi.h"
#include "psvbi-codes.cpp"

//#define PSVBI_SINGLE_FIND

#define CLEAR(x) memset (&(x), 0, sizeof (x))

static unsigned char medianna(unsigned char b1, unsigned char b2, unsigned char b3)
{
    #define MID(B1,B2,B3) if(((B1<B2)&&(B1>B3))||((B1>B2)&&(B1<B3))) return B1;

    MID(b1,b2,b3)
    MID(b2,b3,b1)
    return b3;
};

unsigned char rgb2y(unsigned int rgb)
{
        return
        (
                ((rgb & 0x00FF0000) >> 16) * ((unsigned int)(256*0.299))       // R
                +
                ((rgb & 0x0000FF00) >> 8) * ((unsigned int)(256*0.587))        // G
                +
                ((rgb & 0x000000FF) >> 0) * ((unsigned int)(256*0.114))        // B
        ) >> 8;
};



// encode packet
int psvbi_encode(struct psvbi_block* block)
{
    int i;

    // make table resolution and put it to block    
    for(i=0;i<PSVBI_DATA_LEN;i++)
    {
        block->encoded[2*i + 0] = line_block_code[ (block->decoded[i] >> 0)&0xf ][8];
        block->encoded[2*i + 1] = line_block_code[ (block->decoded[i] >> 4)&0xf ][8];
/* printf
(
"block->decoded[i] = %.2X /line_block_code[ (block->decoded[i] >> 0)&0xf ][8] == %.2X / line_block_code[ (block->decoded[i] >> 4)&0xf ][8] = %.2X\n",
(unsigned int)block->decoded[i],
(unsigned int)line_block_code[ (block->decoded[i] >> 0)&0xf ][8],
(unsigned int)line_block_code[ (block->decoded[i] >> 4)&0xf ][8]
); */
    };

    return 1;
};


// decode packet
int psvbi_decode(struct psvbi_block* block)
{
    int i;

    unsigned char *lo,*hi;

    // clear attributes
    block->errors = 0;
    block->corrections = 0;
    
    // make table resolution for decoding block
    for(i=0;i<PSVBI_DATA_LEN;i++)
    {
        // find description according to table
        lo = line_block_decode[ block->encoded[2*i + 0] ];
        hi = line_block_decode[ block->encoded[2*i + 1] ];

        // calc errors for thing byte
        block->decoded[i] = (*(lo + 2) << 0) | (*(hi + 2) << 4);
        
        // calc global errors counter
        block->errors += (block->decode_errors[i] = *hi + *lo );

        // calc corrections in a case of error not happen        
        block->decode_corrections[i] = (block->decode_errors[i] == 0)? ( *(hi + 1) + *(lo + 1) ) : 0;

        // calc global corrections counter
        block->corrections += block->decode_corrections[i];
    };

    // return OK in a case of no errors happens
    return (block->errors)?0:1;
};


// save data to planar Y
int psvbi_save_Y(struct psvbi_block* block,void* data,int offset,unsigned char level)
{
    int i;
    unsigned char *ys = (unsigned char*)data + offset;
    unsigned char r;
    int bit,byte,pos;
    
    for(i=0;i<PSVBI_PACKET_LEN_UNPACKED;i++)
    {
        // calc source and destination bit position
        byte = i>>3; bit = i%8; pos = bit*PSVBI_PACKET_LEN_PACKED + byte;

//fprintf(stderr,"i=%d,byte=%d,bit=%d,pos=%d,local_offset=%d,global_offset=%d",i,byte,bit,pos,PSVBI_DENSITY*pos,offset);

        // resolv source bit
        r = ( block->encoded[byte] & (1<<bit) )?1:0;
//fprintf(stderr,"\tsaving ...");        
        // put data into destination Y planar buffer
        ys[PSVBI_DENSITY*pos + 0] = (r)?level:0;
        ys[PSVBI_DENSITY*pos + 1] = ys[PSVBI_DENSITY*pos + 0];
        ys[PSVBI_DENSITY*pos + 2] = ys[PSVBI_DENSITY*pos + 0];
//fprintf(stderr,"OK\n");
    };

    return 1;
};


// save data to RGB struct
int psvbi_save_RGB(struct psvbi_block* block,void* data,int offset,unsigned char level)
{
    int i;
    unsigned int *rgbs = (unsigned int*)((unsigned char*)data + 4*offset);
    unsigned char r;
    int bit,byte,pos;
    
    unsigned int l = (((unsigned int)level)<<16) | (((unsigned int)level)<<8) |  (((unsigned int)level)<<0);
    
    for(i=0;i<PSVBI_PACKET_LEN_UNPACKED;i++)
    {
        // calc source and destination bit position
        byte = i>>3; bit = i%8; pos = bit*PSVBI_PACKET_LEN_PACKED + byte;
        
        // resolv source bit
        r = ( block->encoded[byte] & (1<<bit) )?1:0;
        
        // put data into destination Y planar buffer
        rgbs[PSVBI_DENSITY*pos + 0] = (r)?l:0;
        rgbs[PSVBI_DENSITY*pos + 1] = rgbs[PSVBI_DENSITY*pos + 0];
        rgbs[PSVBI_DENSITY*pos + 2] = rgbs[PSVBI_DENSITY*pos + 0];
    };

    return 1;
};                                                                                                                                              
                                                                                                                                               
// load data from planar Y
int psvbi_load_Y(struct psvbi_block* block,void* data,int offset,unsigned char level)
{
    int i;
    unsigned char *ys = (unsigned char*)data + offset;
    unsigned char r;
    int bit,byte,pos;

    for(i=0;i<PSVBI_PACKET_LEN_UNPACKED;i++)
    {
        // resolv "middle" value
        r = medianna(*(ys + i*PSVBI_DENSITY + 0), *(ys + i*PSVBI_DENSITY + 1), *(ys + i*PSVBI_DENSITY + 2) );
        
        // resolv "level-depend" value
        r = (r > level)?1:0;

        // calc bit position accoring to persmutation rule        
        bit = (unsigned char)(i/PSVBI_PACKET_LEN_PACKED);
        byte = (unsigned char)(i%PSVBI_PACKET_LEN_PACKED);

        // put into position depends on permutation rule
        if(r)
            block->encoded[byte] |= (1<<bit);		// set bit
        else
            block->encoded[byte] &= 0xFF ^ (1<<bit);	// clear bit
        
    };
    return 1;
};

// load data from RGB struct
int psvbi_load_RGB(struct psvbi_block* block,void* data,int offset,unsigned char level)
{
    int i;
    unsigned int *rgbs = (unsigned int*)((unsigned char*)data + 4*offset);
    unsigned char r;
    int bit,byte,pos;

    for(i=0;i<PSVBI_PACKET_LEN_UNPACKED;i++)
    {
        // resolv "middle" value
        r = medianna( rgb2y(rgbs[i*PSVBI_DENSITY + 0]), rgb2y(rgbs[i*PSVBI_DENSITY + 1]), rgb2y(rgbs[i*PSVBI_DENSITY + 2]) );
        
        // resolv "level-depend" value
        r = (r > level)?1:0;

        // calc bit position accoring to persmutation rule        
        bit = (unsigned char)(i/PSVBI_PACKET_LEN_PACKED);
        byte = (unsigned char)(i%PSVBI_PACKET_LEN_PACKED);

        // put into position depends on permutation rule
        if(r)
            block->encoded[byte] |= (1<<bit);		// set bit
        else
            block->encoded[byte] &= 0xFF ^ (1<<bit);	// clear bit
        
    };
    return 1;
};

int psvbi_find_Y(struct psvbi_block* block, void* data, int h_from,int h_to, int v_from, int v_to,int line_len,unsigned char level)
{
    int i,j;
    struct psvbi_block best_block;
    struct psvbi_block error_block;
    int best_block_found = 0;

    // set correction field to maximum
    CLEAR(best_block);
    CLEAR(error_block);
    best_block.corrections = 0xFF;
    error_block.errors = 0xFF;

    for(j=v_from;j<v_to;j++)
        for(i=h_from;i<h_to;i++)
        {	
            // probe struct
            struct psvbi_block try_block;
            CLEAR(try_block);
          
            // load data
            psvbi_load_Y(&try_block,data,j*line_len + i,level);
          
            // decoder
            if (psvbi_decode(&try_block))
            {
                // set flag data found 
                best_block_found = 1;
              
                // check if correction less then current
                if(try_block.corrections < best_block.corrections)
                {
                    // update best block
                    best_block = try_block;
                    
                    // save data about search position
                    best_block._h = i;
                    best_block._v = j;
                };
		
#ifdef PSVBI_SINGLE_FIND
		*block = best_block;
		return best_block_found;
#endif	// PSVBI_SINGLE_FIND

            }
	    else
	    {
		if(try_block.errors < error_block.errors)
		    error_block = try_block;
	    };
        };

    // check if we need to update given block
    if (best_block_found)
        *block = best_block;
    else
	*block = error_block;
        
    // return result
    return best_block_found;
};

int psvbi_find_RGB(struct psvbi_block* block, void* data, int h_from,int h_to, int v_from, int v_to,int line_len, unsigned char level)
{
    int i,j;
    struct psvbi_block best_block;
    struct psvbi_block error_block;
    int best_block_found = 0;
    

    // set correction field to maximum
    CLEAR(best_block);
    CLEAR(error_block);
    best_block.corrections = 0xFF;
    error_block.errors = 0xFF;

    for(j=v_from;j<v_to;j++)
        for(i=h_from;i<h_to;i++)
        {	
            // probe struct
            struct psvbi_block try_block;
            CLEAR(try_block);
          
            // load data
            psvbi_load_RGB(&try_block,data,j*line_len + i,level);
          
            // decoder
            if (psvbi_decode(&try_block))
            {
                // set flag data found 
                best_block_found = 1;
              
                // check if correction less then current
                if(try_block.corrections < best_block.corrections)
                {
                    // update best block
                    best_block = try_block;
                    
                    // save data about search position
                    best_block._h = i;
                    best_block._v = j;
                };
            }
	    else
	    {
		if(try_block.errors < error_block.errors)
		    error_block = try_block;
	    };
        };

    // check if we need to update given block
    if (best_block_found)
        *block = best_block;
    else
	*block = error_block;
        
    // return result
    return best_block_found;
};

void psvbi_dump_block(struct psvbi_block* block)
{
    int i;
    
    printf("/* ------------- struct psvbi_block DUMP ----------------------\n");
    printf("GLOBAL: -> errors: %d, corrections: %d, _h: %d, _v: %d \n",(unsigned int)block->errors,(unsigned int)block->corrections,(unsigned int)block->_h,(unsigned int)block->_v);
    printf("HEX DATA1:  -> ");
    for(i=0;i<PSVBI_DATA_LEN;i++)
        printf("%3.2X",(unsigned int)block->decoded[i]);
    printf("\n");
    printf("HEX DATA2:  -> ");
    for(i=0;i<PSVBI_PACKET_LEN_PACKED;i++)
        printf("%3.2X",(unsigned int)block->encoded[i]);
    printf("\n");
    printf("ERRORS:     -> ");
    for(i=0;i<PSVBI_DATA_LEN;i++)
        printf("%3.1d",(unsigned int)block->decode_errors[i]);
    printf("\n");
    printf("CORRECTIONS:-> ");
    for(i=0;i<PSVBI_DATA_LEN;i++)
        printf("%3.1d",(unsigned int)block->decode_corrections[i]);
    printf("\n");
    printf("LITERALS:   -> ");
    for(i=0;i<PSVBI_DATA_LEN;i++)
        printf("%3.1c",block->decoded[i]);
    printf("\n");
    printf("-------------------------------------------------------------- */ \n");        
    
};
