#ifndef _PSVBI_H
#define _PSVBI_H

#ifndef LOG_DECODER_ERRORS
#define STOP_DECODER_ON_ERROR
#endif // LOG_DECODER_ERRORS

#define PSVBI_PACKET_LEN_UNPACKED	224	/* lenght of unpacked coded data packet */
#define PSVBI_PACKET_LEN_PACKED		28	/* lenght of packed coded data packet */
#define PSVBI_DATA_LEN			14	/* lenght of decoded data packet */
#define PSVBI_DENSITY			3	/* number of samples to code 1 bit */

struct psvbi_block
{
    // in/out datas
    unsigned char decoded[PSVBI_DATA_LEN];		// data block to coded/restored
    unsigned char encoded[PSVBI_PACKET_LEN_PACKED];	// coded data block

    // statistic
    unsigned char decode_errors[PSVBI_DATA_LEN];	// errors numbers for bytes
    unsigned char decode_corrections[PSVBI_DATA_LEN];	// number of corrected bit
    int errors;						// total errors count during decoding
    int corrections;					// total count of corrected bits

    // decoding/coding controls
    int _h;						// horizontal and vertical positions of
    int _v;						// found block
/*    
    int state;						// state-machine controls
    void* data;						// input - output data 
    int offset;						// offset from begining
*/
};


// encode packet
int psvbi_encode(struct psvbi_block* block);
// save data to planar Y
int psvbi_save_Y(struct psvbi_block* block,void* data,int offset,unsigned char level);
// save data to RGB struct
int psvbi_save_RGB(struct psvbi_block* block,void* data,int offset,unsigned char level);


// decode packet
int psvbi_decode(struct psvbi_block* block);
// load data from planar Y
int psvbi_load_Y(struct psvbi_block* block,void* data,int offset,unsigned char level);
// load data from RGB struct
int psvbi_load_RGB(struct psvbi_block* block,void* data,int offset, unsigned char level);

// try to find block in proposed range
int psvbi_find_Y(struct psvbi_block* block, void* data, int h_from,int h_to, int v_from, int v_to,int line_len, unsigned char level);
int psvbi_find_RGB(struct psvbi_block* block, void* data, int h_from,int h_to, int v_from, int v_to,int line_len, unsigned char level);

// debug features
void psvbi_dump_block(struct psvbi_block* block);

#endif // _PSVBI_H
