#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>              /* low-level i/o */
#include <unistd.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <asm/types.h>          /* for videodev2.h */
#include <linux/videodev2.h>

#include "v4l2device.h"
#include "psvbid.h"


#define CLEAR(x) memset (&(x), 0, sizeof (x))

#define FRAMEBUFFER_LEN 5


static int fd;					// video device descriptor
static char dev_name[1024] = "/dev/video0";		// video device name
static void* framebuffer[FRAMEBUFFER_LEN];
static int framebuffer_len[FRAMEBUFFER_LEN];
static v4l2device_fgcb callback_frame_process = NULL;
static int exit_flag = 0;

/*
    SETUP PROCS
*/
static void open_device()
{
    struct stat st;
                                                                                                                                               
    if (-1 == stat (dev_name, &st))
    {
        fprintf (stderr, "Cannot identify '%s': %d, %s\n", dev_name, errno, strerror (errno));
        exit (EXIT_FAILURE);
    };
                                                                                                                                               
    if (!S_ISCHR (st.st_mode))
    {
        fprintf (stderr, "%s is no device\n", dev_name);
        exit (EXIT_FAILURE);
    }
                                                                                                                                               
    fd = open (dev_name, O_RDWR /* required */ | O_NONBLOCK, 0);
                                                                                                                                               
    if (-1 == fd)
    {
        fprintf (stderr, "Cannot open '%s': %d, %s\n", dev_name, errno, strerror (errno));
        exit (EXIT_FAILURE);
    };
};

static void setup_device()
{
    // framebuffset params
    struct v4l2_format fb_setup;
    CLEAR(fb_setup);
    fb_setup.type		 = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    fb_setup.fmt.pix.width       = PSVBID_FRAME_WIDTH;
    fb_setup.fmt.pix.height      = PSVBID_FRAME_HEIGHT;
    fb_setup.fmt.pix.pixelformat = V4L2_PIX_FMT_GREY;
/*    fb_setup.fmt.pix.pixelformat = V4L2_PIX_FMT_YUV422P; */
    fb_setup.fmt.pix.field       = V4L2_FIELD_INTERLACED;

    // video input
    int video_input = PSVBID_INPUT_NO;

    // television standart
    __u64 video_standart = V4L2_STD_PAL_D ;

    // setup VIDIOC_S_FMT
    printf("VIDIOC_S_FMT: V4L2_BUF_TYPE_VIDEO_CAPTURE,i%dx%d,V4L2_PIX_FMT_YUV422P ...",fb_setup.fmt.pix.width,fb_setup.fmt.pix.height);
    if(-1 == ioctl(fd, VIDIOC_S_FMT, &fb_setup) )
    {
        fprintf (stderr, "Error(VIDIOC_S_FMT): %d, %s\n", errno, strerror (errno));
        exit (EXIT_FAILURE);
    }
    printf("OK [imagesize is: %d bytes]\n", fb_setup.fmt.pix.sizeimage);

    // setup VIDIOC_S_INPUT
    printf("VIDIOC_S_INPUT -> 1 [Composite1]...");
    if(-1 == ioctl(fd, VIDIOC_S_INPUT, &video_input) )
    {
        fprintf (stderr, "Error(VIDIOC_S_INPUT): %d, %s\n", errno, strerror (errno));
        exit (EXIT_FAILURE);
    }
    printf("OK\n");

    // setup VIDIOC_S_STD
    printf("VIDIOC_S_STD -> V4L2_STD_PAL_D...");
    if(-1 == ioctl(fd, VIDIOC_S_STD, &video_standart) )
    {
        fprintf (stderr, "Error(VIDIOC_S_STD): %d, %s\n", errno, strerror (errno));
        exit (EXIT_FAILURE);
    }
    printf("OK\n");
};


static void info_device()
{
    int i;
    
    // VIDIOC_QUERYCAP
    struct v4l2_capability cap;
    if(-1 == ioctl(fd, VIDIOC_QUERYCAP, &cap) )
    {
        fprintf (stderr, "Error(VIDIOC_QUERYCAP): %d, %s\n", errno, strerror (errno));
        exit (EXIT_FAILURE);
    }
    printf("CAPABILITIES:\n");
    printf("\t%25s: %s\n","V4L2_CAP_VIDEO_CAPTURE", (cap.capabilities & V4L2_CAP_VIDEO_CAPTURE)?"YES":"NO");
    printf("\t%25s: %s\n","V4L2_CAP_VIDEO_OUTPUT", (cap.capabilities & V4L2_CAP_VIDEO_OUTPUT)?"YES":"NO");
    printf("\t%25s: %s\n","V4L2_CAP_VIDEO_OVERLAY", (cap.capabilities & V4L2_CAP_VIDEO_OVERLAY)?"YES":"NO");
    printf("\t%25s: %s\n","V4L2_CAP_VBI_CAPTURE", (cap.capabilities & V4L2_CAP_VBI_CAPTURE)?"YES":"NO");
    printf("\t%25s: %s\n","V4L2_CAP_VBI_OUTPUT", (cap.capabilities & V4L2_CAP_VBI_OUTPUT)?"YES":"NO");
    printf("\t%25s: %s\n","V4L2_CAP_RDS_CAPTURE", (cap.capabilities & V4L2_CAP_RDS_CAPTURE)?"YES":"NO");
    printf("\t%25s: %s\n","V4L2_CAP_TUNER", (cap.capabilities & V4L2_CAP_TUNER)?"YES":"NO");
    printf("\t%25s: %s\n","V4L2_CAP_AUDIO", (cap.capabilities & V4L2_CAP_AUDIO)?"YES":"NO");
    printf("\t%25s: %s\n","V4L2_CAP_READWRITE", (cap.capabilities & V4L2_CAP_READWRITE)?"YES":"NO");
    printf("\t%25s: %s\n","V4L2_CAP_ASYNCIO", (cap.capabilities & V4L2_CAP_ASYNCIO)?"YES":"NO");
    printf("\t%25s: %s\n","V4L2_CAP_STREAMING", (cap.capabilities & V4L2_CAP_STREAMING)?"YES":"NO");
//    printf("\t%s: %s\n","__", (cap.capabilities & __)?"YES","NO");
    
    // VIDIOC_ENUM_FMT
    // enumerate image format supported
    struct v4l2_fmtdesc images_fmt;
    printf("VIDIOC_ENUM_FMT:\n");
    for(i = 0; i < 32; i++)
    {
	CLEAR(images_fmt);
        images_fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	images_fmt.index = i;
    
	if(-1 != ioctl(fd, VIDIOC_ENUM_FMT, &images_fmt) )
	{
	    printf("\t%3d: %.8X\t%s\n",images_fmt.index, images_fmt.pixelformat,images_fmt.description);
	};	    
    };

    // VIDIOC_ENUMINPUT
    // enumerate inputs supported
    struct v4l2_input video_input;
    printf("VIDIOC_ENUMINPUT:\n");
    for(i = 0; i < 32; i++)
    {
	CLEAR(video_input);
	video_input.index = i;
    
	if(-1 != ioctl(fd, VIDIOC_ENUMINPUT, &video_input) )
	{
	    printf("\t%3d: %s\n", video_input.index, video_input.name);
	};	    
    };
};

static void init_buffers()
{
    int i;
    
    struct v4l2_requestbuffers req;
    CLEAR (req);
    req.count	= FRAMEBUFFER_LEN;
    req.type	= V4L2_BUF_TYPE_VIDEO_CAPTURE;
    req.memory	= V4L2_MEMORY_MMAP;

    // VIDIOC_REQBUFS
    printf("VIDIOC_REQBUFS: %d buffers ...", FRAMEBUFFER_LEN);
    if (-1 == ioctl (fd, VIDIOC_REQBUFS, &req))
    {
        fprintf (stderr, "Error(VIDIOC_REQBUFS): %d, %s\n", errno, strerror (errno));
        exit (EXIT_FAILURE);
    }
    printf("OK [req.count = %d]\n", req.count);

    // check if requested buffers count is correct
    
    // VIDIOC_QUERYBUF
    printf("VIDIOC_QUERYBUF: mapping ...");
    for(i = 0; i<FRAMEBUFFER_LEN; i++)
    {
        struct v4l2_buffer buf;
        CLEAR (buf);
        buf.type	= V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory	= V4L2_MEMORY_MMAP;
        buf.index	= i;

        if (-1 == ioctl (fd, VIDIOC_QUERYBUF, &buf))
        {
            fprintf (stderr, "Error(VIDIOC_QUERYBUF): %d, %s\n", errno, strerror (errno));
            exit (EXIT_FAILURE);
        }
        framebuffer_len[i] = buf.length;
        framebuffer[i] = mmap 
        (
            NULL /* start anywhere */,
            buf.length,
            PROT_READ | PROT_WRITE /* required */,
            MAP_SHARED /* recommended */,
            fd, buf.m.offset
        );
        if (MAP_FAILED == framebuffer)
        {
            fprintf (stderr, "Error(mmap): %d, %s\n", errno, strerror (errno));
            exit (EXIT_FAILURE);
        }
    };
    printf("OK\n");
					
};

static void read_framebuffer(int no)
{
    int i;
    struct v4l2_buffer buf;
    CLEAR (buf);
                                                                                                                                               
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;

    // enqueue frame                                                                                                                                               
    if (-1 == ioctl (fd, VIDIOC_DQBUF, &buf)) 
    {
        fprintf (stderr, "Error(VIDIOC_DQBUF): %d, %s\n", errno, strerror (errno));
        exit (EXIT_FAILURE);
    };
    
//    assert (buf.index < FRAMEBUFFER_LEN);
    
    // image here
    if(callback_frame_process)
        callback_frame_process(no, buf.index,  framebuffer[buf.index]);

    // dequeue frame                                                                                                                                               
    if (-1 == ioctl (fd, VIDIOC_QBUF, &buf))
    {
        fprintf (stderr, "Error(VIDIOC_DQBUF): %d, %s\n", errno, strerror (errno));
        exit (EXIT_FAILURE);
    };
};

static void read_framebuffer_loop()
{
    struct v4l2_buffer buf;
    int i = 0;

    // START
    enum v4l2_buf_type type;
    // enqueue all mapped buffers
    for(i=0; i< FRAMEBUFFER_LEN ; i++)
    {
        CLEAR (buf);
        buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        buf.memory      = V4L2_MEMORY_MMAP;
        buf.index       = i;
        if (-1 == ioctl (fd, VIDIOC_QBUF, &buf))
        {
            fprintf (stderr, "Error(VIDIOC_QBUF): %d, %s\n", errno, strerror (errno));
            exit (EXIT_FAILURE);
        };
    };

    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (-1 == ioctl (fd, VIDIOC_STREAMON, &type))
    {
        fprintf (stderr, "Error(VIDIOC_STREAMON): %d, %s\n", errno, strerror (errno));
        exit (EXIT_FAILURE);
    };

    printf("Starting main loop....\n");

    // LOOP
    while(!(exit_flag))
    {
        fd_set fds;
        struct timeval tv;
        int r;

        FD_ZERO (&fds);
        FD_SET (fd, &fds);
        
        /* Timeout. */
        tv.tv_sec = 2;
        tv.tv_usec = 0;

        if ( (r = select (fd + 1, &fds, NULL, NULL, &tv)) > 0)
            read_framebuffer(i++);
        else
            fprintf (stderr, "Error/Timeout(select): %d, %s, r=%d\n", errno, strerror (errno),r);
    };

    // STOP
    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (-1 == ioctl (fd, VIDIOC_STREAMOFF, &type))
    {
        fprintf (stderr, "Error(VIDIOC_STREAMOFF): %d, %s\n", errno, strerror (errno));
        exit (EXIT_FAILURE);
    };

};

static void release_buffers()
{
    int i;
    
    for(i=0;i<FRAMEBUFFER_LEN;i++)
    {
        if (-1 == munmap (framebuffer[i], framebuffer_len[i]))
        {
            fprintf (stderr, "Error(munmap): %d, %s\n", errno, strerror (errno));
            exit (EXIT_FAILURE);
        };
    };

};

void video_open(int show_compats, char* dev)
{
    if(dev)
	strncpy(dev_name,dev,1024);

    open_device();

    if (show_compats)    
        info_device();
};

void video_stop()
{
    exit_flag = 1;
};

void video_start(v4l2device_fgcb callback_frame_proc)
{
    // setup and videodevice
    setup_device();
    init_buffers();

    // init callback proc
    callback_frame_process = (v4l2device_fgcb)callback_frame_proc;
    
    // start loop
    read_framebuffer_loop();

    return;


    // close/unmap buffers
    release_buffers();

    // close device
    close(fd);
};

