#include <Filter.h>
#include <ScriptInterpreter.h>
#include <ScriptValue.h>
#include "resource.h"
#include <stdio.h>

#define	PSVBID_SAMPLES			720		/*	[pixels] */
#define PSVBID_LEVEL_SETUP		0x80
#define PSVBID_PIXEL_OFFSET 24

#include "../decoder/psvbi.h"

#define MAX_BLOCKS	4

/* --------------------------------------------------------------------------

  Filter's function definition

-------------------------------------------------------------------------- */
int _RunProc(const FilterActivation *fa, const FilterFunctions *ff);
int _StartProc(FilterActivation *fa, const FilterFunctions *ff);
int _EndProc(FilterActivation *fa, const FilterFunctions *ff);
long _ParamProc(FilterActivation *fa, const FilterFunctions *ff);
int _ConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd);
void _StringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str);
void _ScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc);
bool _FssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen);


/* --------------------------------------------------------------------------

  Runtime and init structs

-------------------------------------------------------------------------- */
struct FilterData 
{
	FILE* logfile;					// file handler for logging

	long frame_start;				// start encoding from frame no:

	long frames_number;				// number of frames to encode 

	int op[MAX_BLOCKS][3];			// encoding provides in 4 same blocks:
									//		encode_op[j][0] - encode flag
									//		encode_op[j][1] - offset
									//		encode_op[j][2] - line
									//	 j - block number

	int level;

	unsigned char data[PSVBI_DATA_LEN];

	int	_setuped;					// flag indicating is this first
									// setup or not.

};

static struct FilterData FilterDataDefault = 
{
	// logfile
	NULL,

	// start, dur
	0,10,

	// encode op
	{
		{1,PSVBID_PIXEL_OFFSET,0},
		{1,PSVBID_PIXEL_OFFSET,1},
		{0,PSVBID_PIXEL_OFFSET,2},
		{0,PSVBID_PIXEL_OFFSET,3}
	},

	// levels
	PSVBID_LEVEL_SETUP,
	{0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0x0B,0x0C,0x0D,0x0E},
	1
};


ScriptFunctionDef _func_defs[]=
{
    { (ScriptFunctionPtr)_ScriptConfig, "Config", "0i." },
    { NULL },
};

CScriptObject _obj=
{
    NULL,
	_func_defs
};

struct FilterDefinition filterDef = {

    NULL,									//	struct FilterDefinition *next, *prev;
	NULL,									//
	NULL,									//	FilterModule *module;
    "pseudovbi",							//	char *name;
    "encode data into image line",			//	char *desc;
    "verem",								//	char *maker;
    NULL,									//	void *private_data;
    sizeof(FilterData),						//	int inst_data_size;

    NULL,									//	FilterInitProc initProc;
    NULL,									//	FilterDeinitProc deinitProc;

    _RunProc,								//	FilterRunProc runProc;
    _ParamProc,								//	FilterParamProc paramProc;
    _ConfigProc,							//	FilterConfigProc configProc;
    _StringProc,							//	FilterStringProc stringProc
    _StartProc,								//	FilterStartProc startProc;
    _EndProc,								//	FilterEndProc endProc;

    &_obj,									//	CScriptObject *script_obj;
    _FssProc,								//	FilterScriptStrProc fssProc;

};

/* --------------------------------------------------------------------------

  Exports definition

-------------------------------------------------------------------------- */

extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
extern "C" void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff);
static FilterDefinition *filter_definition;



/* --------------------------------------------------------------------------

  Exports definition

-------------------------------------------------------------------------- */
int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) 
{
    if (!(filter_definition = ff->addFilter(fm, &filterDef, sizeof(FilterDefinition))))
        return 1;

	vdfd_ver = VIRTUALDUB_FILTERDEF_VERSION;
	vdfd_compat = VIRTUALDUB_FILTERDEF_COMPATIBLE;

    return 0;
}

void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff) 
{
    ff->removeFilter(filter_definition);
}


/* --------------------------------------------------------------------------

  RUNTIME PROCS

-------------------------------------------------------------------------- */

int _StartProc(FilterActivation *fa, const FilterFunctions *ff) 
{
    FilterData *filter_data = (FilterData *)fa->filter_data;
    return 0;
}

int _EndProc(FilterActivation *fa, const FilterFunctions *ff)
{
    return 0;
}

/* --------------------------------------------------------------------------

  IMAGE PROCESSING ROUTINE

-------------------------------------------------------------------------- */

#define LINE_PTR(LINE)												\
	(																\
		(unsigned long *)											\
		(															\
			(unsigned char*)fa->dst.data							\
			+														\
			(fa->dst.modulo + sizeof(Pixel32)*fa->dst.w)*fa->dst.h	\
			-														\
			(fa->dst.w)*sizeof(Pixel32)*((LINE) + 1)				\
		)															\
	)

int _RunProc(const FilterActivation *fa, const FilterFunctions *ff)
{
    FilterData *filter_data = (FilterData *)fa->filter_data;
	int i;

	// copy data to dst buffer
	memcpy(fa->dst.data,fa->src.data, (fa->src.modulo + sizeof(Pixel32)*fa->src.w)*fa->src.h);

	// check tha range of frames
	long frames_distance = fa->pfsi->lCurrentSourceFrame - filter_data->frame_start;
	if
	(
		(frames_distance >= 0)
		&&
		(frames_distance < filter_data->frames_number)
	)
	{
		// encode datas
		struct psvbi_block block;

		// copy data to block
		memcpy(block.decoded, filter_data->data, PSVBI_DATA_LEN);

		// encode data
		psvbi_encode(&block);

		// put data
		for(i=0;i<4;i++)
			if(filter_data->op[i][0])
			{
				// clear line
				memset
				(
					LINE_PTR(filter_data->op[i][2]),
					0,
					PSVBID_SAMPLES*sizeof(Pixel32)
				);
				psvbi_save_RGB
				(
					&block,
					LINE_PTR(filter_data->op[i][2]),
					filter_data->op[i][1],
					filter_data->level
				);
			};

	};

    return 0;
}

///////////////////////////////////////////////////////////////////////////

long _ParamProc(FilterActivation *fa, const FilterFunctions *ff) {

    FilterData *filter_data = (FilterData *)fa->filter_data;

    return FILTERPARAM_SWAP_BUFFERS;
}


BOOL CALLBACK _ConfigDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam) 
{
#define SET_DLG_INT(ID,VAL)			\
	ltoa(VAL, buf, 10);				\
	SetDlgItemText(hdlg, ID, buf);

#define GET_DLG_INT(ID,VAL)						\
	GetDlgItemText(hdlg, ID, buf, sizeof(buf));	\
	VAL = atol(buf);
	
	int i;
	char buf[128];
	static encode_op_fields[4][3] = 
	{
		{ IDC_BLOCK_1, IDC_EDIT_POS_X1,	IDC_EDIT_POS_Y1},
		{ IDC_BLOCK_2, IDC_EDIT_POS_X2,	IDC_EDIT_POS_Y2},
		{ IDC_BLOCK_3, IDC_EDIT_POS_X3,	IDC_EDIT_POS_Y3},
		{ IDC_BLOCK_4, IDC_EDIT_POS_X4,	IDC_EDIT_POS_Y4}
	};

    FilterData *filter_data = (FilterData *)GetWindowLong(hdlg, DWL_USER);

    switch(msg) 
	{

		case WM_INITDIALOG:
			// setup data
            SetWindowLong(hdlg, DWL_USER, lParam);

			// request filter data pointer
			filter_data = (FilterData *)lParam;

			// check if it's first diag
			if(!(filter_data->_setuped))
			{
				// save file pointer
				void* temp = filter_data->logfile;
				*filter_data = FilterDataDefault;
				filter_data->logfile = (FILE*)temp;
			};

			// blocks
			for(i=0;i<MAX_BLOCKS;i++)
			{
				CheckDlgButton(hdlg, encode_op_fields[i][0], filter_data->op[i][0]?BST_CHECKED:BST_UNCHECKED);
				SET_DLG_INT(encode_op_fields[i][1],filter_data->op[i][1]);
				SET_DLG_INT(encode_op_fields[i][2],filter_data->op[i][2]);
			};

			// levels
			SET_DLG_INT(IDC_EDIT_LEVEL,	filter_data->level);

			// frames coords
			SET_DLG_INT(IDC_EDIT_FRAME_START, filter_data->frame_start);
			SET_DLG_INT(IDC_EDIT_FRAMES_NUMBER, filter_data->frames_number);

			// data block
			sprintf
			(
				buf,"%3.2X%3.2X%3.2X%3.2X%3.2X%3.2X%3.2X%3.2X%3.2X%3.2X%3.2X%3.2X%3.2X%3.2X",
				(int)filter_data->data[0],(int)filter_data->data[1],(int)filter_data->data[2],
				(int)filter_data->data[3],(int)filter_data->data[4],(int)filter_data->data[5],
				(int)filter_data->data[6],(int)filter_data->data[7],(int)filter_data->data[8],
				(int)filter_data->data[9],(int)filter_data->data[10],(int)filter_data->data[11],
				(int)filter_data->data[12],(int)filter_data->data[13]
			);
			SetDlgItemText(hdlg, IDC_EDIT_DATA_BLOCK, buf);


            return TRUE;

        case WM_COMMAND:
            switch(LOWORD(wParam)) 
			{
				case IDOK:

					// request encode params
					for(i=0;i<MAX_BLOCKS;i++)
					{
						filter_data->op[i][0] = !!IsDlgButtonChecked(hdlg, encode_op_fields[i][0]);
						GET_DLG_INT(encode_op_fields[i][1],filter_data->op[i][1]);
						GET_DLG_INT(encode_op_fields[i][2],filter_data->op[i][2]);
					};

					// levels
					GET_DLG_INT(IDC_EDIT_LEVEL,	filter_data->level);

					// frames coords
					GET_DLG_INT(IDC_EDIT_FRAME_START, filter_data->frame_start);
					GET_DLG_INT(IDC_EDIT_FRAMES_NUMBER, filter_data->frames_number);

					// data block
					GetDlgItemText(hdlg, IDC_EDIT_DATA_BLOCK, buf, sizeof(buf));
					int data[PSVBI_DATA_LEN];
					sscanf
					(
						buf,"%X %X %X %X %X %X %X %X %X %X %X %X %X %X",
						&data[0],&data[1],&data[2],&data[3],&data[4],&data[5],&data[6],
						&data[7],&data[8],&data[9],&data[10],&data[11],&data[12],&data[13]
					);
					for(i=0;i<PSVBI_DATA_LEN;i++)
						filter_data->data[i] = (unsigned char)data[i];
					

					EndDialog(hdlg, 0);
					return TRUE;

				case IDCANCEL:
					EndDialog(hdlg, 1);
					return FALSE;
				}
            break;
    }

    return FALSE;
}

int _ConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd) {
    return DialogBoxParam
	(
		fa->filter->module->hInstModule,
		MAKEINTRESOURCE(IDD_FILTER_DIALOG),
		hwnd,
        _ConfigDlgProc,
		(LPARAM)fa->filter_data
	);
}

///////////////////////////////////////////////////////////////////////////


void _StringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str) 
{
    FilterData *filter_data = (FilterData *)fa->filter_data;
	char buf[1024];

	sprintf
	(
		buf,"%3.2X%3.2X%3.2X%3.2X%3.2X%3.2X%3.2X%3.2X%3.2X%3.2X%3.2X%3.2X%3.2X%3.2X",
		(unsigned int)filter_data->data[0],  (unsigned int)filter_data->data[1], (unsigned int)filter_data->data[2],
		(unsigned int)filter_data->data[3],  (unsigned int)filter_data->data[4], (unsigned int)filter_data->data[5],
		(unsigned int)filter_data->data[6],  (unsigned int)filter_data->data[7], (unsigned int)filter_data->data[8],
		(unsigned int)filter_data->data[9],  (unsigned int)filter_data->data[10],(unsigned int)filter_data->data[11],
		(unsigned int)filter_data->data[12], (unsigned int)filter_data->data[13]
	);

	sprintf
	(
		str,
		" DATA=[%s],FRAME=[%d,+%d]",
		buf,filter_data->frame_start,filter_data->frames_number
	);
}

void _ScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc) 
{
    FilterActivation *fa = (FilterActivation *)lpVoid;
    FilterData *filter_data = (FilterData *)fa->filter_data;
	int p = 0,i,d;

	unsigned short int* data = (unsigned short int*)filter_data->data;

	// frames
	filter_data->frame_start = argv[p++].asInt();
	filter_data->frames_number = argv[p++].asInt();
	filter_data->level = argv[p++].asInt();

	// data
	for(i=0;i<7;i++)
		data[i] = (unsigned short int)argv[p++].asInt();

	// blocks
	for(i=0;i<4;i++)
	{
		d = argv[p++].asInt();
		filter_data->op[i][0] = (0x80000000 & d)?1:0;
		filter_data->op[i][1] = d & 0x0000FFFF;
		filter_data->op[i][2] = (d>>16) & 0x00007FFF;
	};

	// not first (prevent loading of default params
	filter_data->_setuped = 1;
}


bool _FssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen) 
{
    FilterData *filter_data = (FilterData *)fa->filter_data;

	unsigned short int* data = (unsigned short int*)filter_data->data;

#define code_op(N) 	((filter_data->op[N][0])?0x80000000:0) | (filter_data->op[N][1] & 0x0000FFFF) | ((filter_data->op[N][2] << 16) & 0x7FFF0000) 
    _snprintf
	(
		buf,
		buflen,
		"Config(%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d)",

		// frames
		filter_data->frame_start,
		filter_data->frames_number,
		filter_data->level,

		// data
		data[0],data[1],data[2],data[3],data[4],data[5],data[6],data[7],

		// blocks
		code_op(0),code_op(1),code_op(2),code_op(3)
	);


    return true;
}

